#!/usr/bin/env perl
#
# This script, taken from the OpenBSD CVS repository. {{{1
#
# Perl filter to handle the log messages from the checkin of files in
# a directory.  This script will group the lists of files by log
# message, and mail a single consolidated log message at the end of
# the commit.
#
# This file assumes a pre-commit checking program that leaves the
# names of the first and last commit directories in a temporary file.
#
# Contributed by David Hampton <hampton@cisco.com>
#
# hacked greatly by Greg A. Woods <woods@web.net>

# Usage: newlog.pl [-d] [-s] [-M module] [[-m mailto] ...] [-f logfile]
#	-d		- turn on debugging
#	-m mailto	- send mail to "mailto" (multiple)
#	-M modulename	- set module name to "modulename"
#	-f logfile	- write commit messages to logfile too
#	-s		- *don't* run "cvs status -v" for each file
#
#	Configurable options
# }}}1
# 

$MAILER = "/usr/lib/sendmail";
$MAILREPLYTO = "nethack-ko-cvs-commit\@lists.kldp.net";
$MAILDEFAULT = "nethack-ko CVS Commit <nethack-ko-cvs-commit\@lists.kldp.net>";
$MAILFROM = "Changbeom Ahn <bom\@chol.com>";

# find 'cvs' program path {{{1
$cvs = "/usr/bin/cvs";

# Constants (don't change these!) {{{1
#
$STATE_NONE    = 0;
$STATE_CHANGED = 1;
$STATE_ADDED   = 2;
$STATE_REMOVED = 3;
$STATE_LOG     = 4;

$LAST_FILE     = "/tmp/#cvs.lastdir";

$CHANGED_FILE  = "/tmp/#cvs.files.changed";
$ADDED_FILE    = "/tmp/#cvs.files.added";
$REMOVED_FILE  = "/tmp/#cvs.files.removed";
$LOG_FILE      = "/tmp/#cvs.files.log";
$REVS_FILE     = "/tmp/#cvs.files.revs";

$FILE_PREFIX   = "#cvs.files";

# Initialize basic variables {{{1
#
$debug = 0;
# GNOMEó getppid()  error ߻.
$id = getpgrp();	# note, you *must* use a shell which does setpgrp()
$state = $STATE_NONE;

$login = $ENV{'CVS_USERNAME'} || $ENV{'CVS_USER'} || getlogin
	 || (getpwuid($<))[0] || "nobody";

$cvsroot = $ENV{'CVSROOT'};
$do_status = 1;
$modulename = "";

$subject = "no log message";
$tag = "";
# }}}1

#
#	Subroutines
#
# cleanup_tmpfiles {{{1
sub cleanup_tmpfiles {
    local($wd, @files);

    $wd = `pwd`;
    chdir("/tmp") || die("Can't chdir('/tmp')\n");
    opendir(DIR, ".");
    push(@files, grep(/^$FILE_PREFIX\..*\.$id$/, readdir(DIR)));
    closedir(DIR);
    foreach (@files) {
	unlink $_;
    }
    unlink $LAST_FILE . "." . $id;

    chdir($wd);
}

# write_logfile {{{1
sub write_logfile {
    local($filename, @lines) = @_;

    open(FILE, ">$filename") || die("Cannot open log file $filename.\n");
    print FILE join("\n", @lines), "\n";
    close(FILE);
}

# format_names {{{1
sub format_names {
    local($dir, @files) = @_;
    local(@lines);

    if ($dir =~ /^\.\//) {
	$dir = $';
    }
    if ($dir =~ /\/$/) {
	$dir = $`;
    }
    if ($dir eq "") {
	$dir = ".";
    }

    $format = "\t%-" . sprintf("%d", length($dir) > 15 ? length($dir) : 15) . "s%s ";

    $lines[0] = sprintf($format, $dir, ":");

    if ($debug) {
	print STDERR "format_names(): dir = ", $dir, "; files = ", join(":", @files), ".\n";
    }

    $tag_found = "";
    foreach $file (@files) {
	if ($file =~ /^Tag:/) {
	    $tag_found = "1";
	} elsif ($tag_found eq "1") {
	    $tag = "$file";
	    $tag_found = "";
	}
	if (length($lines[$#lines]) + length($file) > 65) {
	    $lines[++$#lines] = sprintf($format, " ", " ");
	}
	$lines[$#lines] .= $file . " ";
    }

    @lines;
}

# format_lists {{{1
sub format_lists {
    local(@lines) = @_;
    local(@text, @files, $lastdir);

    if ($debug) {
	print STDERR "format_lists(): ", join(":", @lines), "\n";
    }
    @text = ();
    @files = ();
    $lastdir = shift @lines;	# first thing is always a directory
    if ($lastdir !~ /.*\/$/) {
	die("Damn, $lastdir doesn't look like a directory!\n");
    }
    foreach $line (@lines) {
	if ($line =~ /.*\/$/) {
	    push(@text, &format_names($lastdir, @files));
	    $lastdir = $line;
	    @files = ();
	} else {
	    push(@files, $line);
	}
    }
    push(@text, &format_names($lastdir, @files));

    @text;
}

# append_names_to_file {{{1
sub append_names_to_file {
    local($filename, $dir, @files) = @_;

    if (@files) {
	open(FILE, ">>$filename") || die("Cannot open file $filename.\n");
	print FILE $dir, "\n";
	print FILE join("\n", @files), "\n";
	close(FILE);
    }
}

# append_revs_to_file {{{1
sub append_revs_to_file {
    local($filename, $src, $oldrev, $newrev, $cwd) = @_;

    open(FILE, ">>$filename") || die("Cannot open file $filename.\n");
    print FILE "$src\%$oldrev\%$newrev\%$cwd\n";
    close(FILE);
}

# read_line {{{1
sub read_line {
    local($line);
    local($filename) = @_;

    open(FILE, "<$filename") || die("Cannot open file $filename.\n");
    $line = <FILE>;
    close(FILE);
    chop($line);
    $line;
}

# read_logfile {{{1
sub read_logfile {
    local(@text);
    local($filename, $leader) = @_;

    open(FILE, "<$filename");
    while (<FILE>) {
	chop;
	push(@text, $leader.$_);
    }
    close(FILE);
    @text;
}

# build_header {{{1
sub build_header {
    local($header);
    local($sec,$min,$hour,$mday,$mon,$year) = localtime(time);
    $header = sprintf("CVSROOT:\t%s\nModule name:\t%s\n",
		      $cvsroot,
		      $modulename);
    if (defined($branch)) {
	$header .= sprintf("Branch: \t%s\n",
		      $branch);
    }
    $header .= sprintf("Changes by:\t%s\t%04d/%02d/%02d %02d:%02d:%02d",
		      $login,
		      $year+1900, $mon+1, $mday,
		      $hour, $min, $sec);
}

# mail_notification() {{{1
sub mail_notification {
    local($name, @text) = @_;
    if ($name eq "") {
	if ($mailother ne "") {
	    $name = $mailother;
	} else {
	    $name = "$MAILDEFAULT";
	}
    }
    open(MAIL, "| $MAILER -t");
    print MAIL "From: $MAILFROM\n";
    print MAIL "To: $name\n";
    print MAIL "Reply-To: $MAILREPLYTO\n";
#    if ($name eq $mailfrom) {
#	print MAIL "Subject: $subject\n";
#    } else {
	print MAIL "Subject: CVS[$modulename,$login] $subject\n";
#    }
    print MAIL "Content-Type: text/plain; charset=euc-kr\n";
    print MAIL "Content-Transfer-Encoding: 8bit\n";
    print MAIL "\n";
    print MAIL join("\n", @text), "\n";
    close(MAIL);
}

# write_commitlog() {{{1
sub write_commitlog {
    local($logfile, @text) = @_;

    open(FILE, ">>$logfile");
    print FILE join("\n", @text), "\n\n";
    close(FILE);
}
# }}}1

#
#	Main Body
#
# Main Body {{{1

# parse command line arguments (file list is seen as one arg)
#
while (@ARGV) {
    $arg = shift @ARGV;

    if ($arg eq '-d') {
	$debug = 1;
	print STDERR "Debug turned on...\n";
    } elsif ($arg eq '-m') {
	if ($mailto) {
	    $mailto = "$mailto, " . shift @ARGV;
	} else {
	    $mailto = shift @ARGV;
	}
    } elsif ($arg eq '-M') {
	$modulename = shift @ARGV;
    } elsif ($arg eq '-s') {
	$do_status = 0;
    } elsif ($arg eq '-f') {
	($commitlog) && die("Too many '-f' args\n");
	$commitlog = shift @ARGV;
    } else {
	($donefiles) && die("Too many arguments!  Check usage.\n");
	$donefiles = 1;

	# NOTE:   loginfo file %{sVv} ؾ Ѵ.
	# mail   cvs diff  ̱  revision  ´.
        @filesets = split(/\s+/, $arg);

	$path_tmp = shift(@filesets);
	push(@files,"$path_tmp");

        foreach (@filesets) {
            chomp;
            my($file,$oldrev,$newrev) = split(/,/);
            $oldrevs{"$file"} = $oldrev;
            $newrevs{"$file"} = $newrev;
            push(@files,"$file");
            push(@revfiles,"$file");
        }
    }
}

# for now, the first "file" is the repository directory being committed,
# relative to the $CVSROOT location
#
@path = split('/', $files[0]);

# XXX there are some ugly assumptions in here about module names and
# XXX directories relative to the $CVSROOT location -- really should
# XXX read $CVSROOT/CVSROOT/modules, but that's not so easy to do, since
# XXX we have to parse it backwards.
#
if ($modulename eq "") {
    $modulename = $path[0];	# I.e. the module name == top-level dir
}
# }}}1

# added by NAMSH
# open mailto file {{{1
$t_mailto = "";
$prev_line = "";
open(FH, "$cvsroot/CVSROOT/modules.mailto");
while(<FH>) {
    chop;
    if ($_ =~ /^#/) {
	next;
    }

    # '\'     
    if ($_ =~ /\\$/) {
	$_ =~ s/\\$//;
	if ($prev_line ne "") {
	    $prev_line = "$prev_line $_";
	} else {
	    $prev_line = $_;
	}
	next;
    }
    if ($prev_line ne "") {
	$_ = "$prev_line $_";
	$prev_line = "";
    }

    @param = split(':', $_);
    $done = 0;
    @mp=split('/', $param[0]);
    for ($i = 0; ; $i++) {
	if ($i > 0 and $mp[$i] eq "") {
	    $modulename = "$param[1]";
	    $t_mailto = "$param[2]";
	    $done = 1;
	    last;
	} elsif ($mp[$i] eq $path[$i]) {
	    next;
	} else {
	    last;
	}
    }
    if ($done) {
	last;
    }
}
close(FH);

# modules.email   ¥ email ּҷ ٲ۴. {{{1
if ($t_mailto ne "") {
    $prev_line = "";
    open(FH, "$cvsroot/CVSROOT/modules.email");
    while (<FH>) {
	chop;
	if ($_ =~ /^#/) {
	    next;
	}

	# '\'     
	if ($_ =~ /\\$/) {
	    $_ =~ s/\\$//;
	    if ($prev_line ne "") {
		$prev_line = "$prev_line $_";
	    } else {
		$prev_line = $_;
	    }
	    next;
	}
	if ($prev_line ne "") {
	    $_ = "$prev_line $_";
	    $prev_line = "";
	}

	#   ϳ space ٲٰ, ',' ڿ ߰
	$_ =~ s/\s+/ /g;
	$_ =~ s/,([^\s])/, \1/g;

	@param = split(':', $_);

	# word substitution  ã ؼ ϴ ̷ ó.
	$winctest = $t_mailto;
	$winctest =~ s/\b$param[0]\b//g;
	$winctest =~ s/\A$param[0]\b//g;
	$winctest =~ s/\b$param[0]\Z//g;
	if ($t_mailto ne $winctest) {
	    $t_mailto =~ s/$param[0]/$param[1]/;
	}
    }
    close(FH);

    # ó  ߰Ǿ  ִ blank, comma  
    $t_mailto =~ s/,\s*$//;
    $t_mailto =~ s/^\s*//;

    $mailother = "$t_mailto";
}
# }}}1
# ended NAMSH

# modules.silent check {{{1
open(FH, "$cvsroot/CVSROOT/modules.silent");
while(<FH>) {
    chomp;
    if($modulename eq $_) {
	$mailto = "$MAILREPLYTO";
	last;
    }
}
close(FH);
if ($commitlog ne "") {
    $commitlog = $cvsroot . "/" . $modulename . "/" . $commitlog unless ($commitlog =~ /^\//);
}
if ($#path == 0) {
    $dir = ".";
} else {
    $dir = join('/', @path[1..$#path]);
}
$dir = $dir . "/";

if ($debug) {
    print STDERR "module - ", $modulename, "\n";
    print STDERR "dir    - ", $dir, "\n";
    print STDERR "path   - ", join(":", @path), "\n";
    print STDERR "files  - ", join(":", @files), "\n";
    print STDERR "id     - ", $id, "\n";
    print STDERR "login  - ", $login, "\n";
    print STDERR "CVS_USERNAME - ", $ENV{'CVS_USERNAME'}, "\n";
}

# Check for a new directory first. {{{1
# This appears with files set as follows:
#
#    files[0] - "path/name/newdir"
#    files[1] - "-"
#    files[2] - "New"
#    files[3] - "directory"
#
if ($files[2] =~ /New/ && $files[3] =~ /directory/) {
    local(@text);

    @text = ();
    push(@text, &build_header());
    push(@text, "");
    push(@text, $files[0]);
    push(@text, "");

    while (<STDIN>) {
	chop;			# Drop the newline
	push(@text, $_);
    }

    &mail_notification($mailto, @text);

    if ($commitlog) {
	&write_commitlog($commitlog, @text);
    }

    exit 0;
}

# Iterate over the body of the message collecting information. {{{1
#
while (<STDIN>) {
    chop;			# Drop the newline

    if (/^Modified Files/) { $state = $STATE_CHANGED; next; }
    if (/^Added Files/)    { $state = $STATE_ADDED;   next; }
    if (/^Removed Files/)  { $state = $STATE_REMOVED; next; }
    if (/^Log Message/)    { $state = $STATE_LOG;     next; }
    if (/^Revision\/Branch/) { /^[^:]+:\s*(.*)/; $branch = $+; next; }

    s/^[ \t\n]+//;		# delete leading whitespace
    s/[ \t\n]+$//;		# delete trailing whitespace

    if ($state == $STATE_CHANGED) { push(@changed_files, split); }
    if ($state == $STATE_ADDED)   { push(@added_files,   split); }
    if ($state == $STATE_REMOVED) { push(@removed_files, split); }
    if ($state == $STATE_LOG)     { push(@log_lines,     $_); }
}

# Strip leading and trailing blank lines from the log message. {{{1
# Also compress multiple blank lines in the body of the message down to a
# single blank line.
#
while ($#log_lines > -1) {
    last if ($log_lines[0] ne "");
    shift(@log_lines);
}
while ($#log_lines > -1) {
    last if ($log_lines[$#log_lines] ne "");
    pop(@log_lines);
}
for ($i = $#log_lines; $i > 0; $i--) {
    if (($log_lines[$i - 1] eq "") && ($log_lines[$i] eq "")) {
	splice(@log_lines, $i, 1);
    }
}

# Check for an import command. {{{1
# This appears with files set as follows:
#
#    files[0] - "path/name"
#    files[1] - "-"
#    files[2] - "Imported"
#    files[3] - "sources"
#
if ($files[2] =~ /Imported/ && $files[3] =~ /sources/) {
    local(@text);

    @text = ();
    push(@text, &build_header());
    push(@text, "");

    push(@text, "Log message:");
    while ($#log_lines > -1) {
	push (@text, "    " . $log_lines[0]);
	shift(@log_lines);
    }

    &mail_notification($mailto, @text);

    if ($commitlog) {
	&write_commitlog($commitlog, @text);
    }

    exit 0;
}

if ($debug) {
    print STDERR "Searching for log file index...";
}

# Find an index to a log file that matches this log message {{{1
#
for ($i = 0; ; $i++) {
    local(@text);

    last if (! -e "$LOG_FILE.$i.$id"); # the next available one
    @text = &read_logfile("$LOG_FILE.$i.$id", "");
    last if ($#text == -1);	# nothing in this file, use it
    last if (join(" ", @log_lines) eq join(" ", @text)); # it's the same log message as another
}
if ($debug) {
    print STDERR " found log file at $i.$id, now writing tmp files.\n";
}

# Spit out the information gathered in this pass. {{{1
#
&append_names_to_file("$CHANGED_FILE.$i.$id", $dir, @changed_files);
&append_names_to_file("$ADDED_FILE.$i.$id",   $dir, @added_files);
&append_names_to_file("$REMOVED_FILE.$i.$id", $dir, @removed_files);
&write_logfile("$LOG_FILE.$i.$id", @log_lines);

# ٲ  mail ߰ {{{1
sort(@revfiles);
foreach (@revfiles) {
    $fullpath = $_;
    if ($oldrevs{$fullpath} eq "NONE") {
	print MAIL "$fullpath: old revision is NONE\n";
    } elsif ($newrevs{$fullpath} eq "NONE") {
	print MAIL "$fullpath: new revision is NONE\n";
    } else {
	push(@changed,$fullpath);
    }
}
foreach (@changed) {
    $f = $_;
    $cwd = `pwd`;
    &append_revs_to_file("$REVS_FILE.$i.$id", $f, $oldrevs{$f}, $newrevs{$f}, $cwd);
}

# Check whether this is the last directory.  If not, quit. {{{1
#
if ($debug) {
    print STDERR "Checking current dir against last dir.\n";
}
$_ = &read_line("$LAST_FILE.$id");

if ($_ ne $cvsroot . "/" . $files[0]) {
    if ($debug) {
	print STDERR sprintf("Current directory %s is not last directory %s.\n", $cvsroot . "/" .$files[0], $_);
    }
    exit 0;
}
if ($debug) {
    print STDERR sprintf("Current directory %s is last directory %s -- all commits done.\n", $files[0], $_);
}

# }}}1
#
#	End Of Commits!
#

exit if($pid = fork);

# This is it.  The commits are all finished.  Lump everything together
# into a single message, fire a copy off to the mailing list, and drop
# it on the end of the Changes file.
#

# Produce the final compilation of the log messages {{{1
#
@text = ();
@status_txt = ();
@diff_txt = ();
push(@text, &build_header());
push(@text, "");

for ($i = 0; ; $i++) {
    last if (! -e "$LOG_FILE.$i.$id"); # we're done them all!

    # log file parse {{{2
    @lines = &read_logfile("$CHANGED_FILE.$i.$id", "");
    if ($#lines >= 0) {
	push(@text, "Modified files:");
	push(@text, &format_lists(@lines));
    }
    @lines = &read_logfile("$ADDED_FILE.$i.$id", "");
    if ($#lines >= 0) {
	push(@text, "Added files:");
	push(@text, &format_lists(@lines));
    }
    @lines = &read_logfile("$REMOVED_FILE.$i.$id", "");
    if ($#lines >= 0) {
	push(@text, "Removed files:");
	push(@text, &format_lists(@lines));
    }
    if ($#text >= 0) {
	push(@text, "");
    }
    @lines = &read_logfile("$LOG_FILE.$i.$id", "");
    if ($#lines >= 0) {
	push(@text, "Log message:");
	push(@text, @lines);
	push(@text, "");
	$tmpsub = "";
	if ($lines[0] ne "") {
	    # ѱ ߸  ϱ    ߶󳽴.
	    # XXX: optimize
	    if (substr($lines[0],49,50) ne "") {
		$tmp = substr($lines[0],0,50);
		while ($tmp) {
		    $pos = index($tmp, " ");
		    if ($pos > 0) {
			if ($tmpsub eq "") {
			    $tmpsub = substr($tmp, 0, $pos);
			} else {
			    $tmpsub = $tmpsub." ".substr($tmp, 0, $pos);
			}
			$pos = $pos + 1;
			$tmp = substr($tmp,$pos,9999);
		    } else {
			last;
		    }
		}
	    } else {
		$subject = substr($lines[0],0,50);
	    }
	}
	if ($tmpsub ne "") {
	    $subject = $tmpsub;
	}
    }

    # generate diff {{{2
    @lines = &read_logfile("$REVS_FILE.$i.$id", "");
    foreach $x (@lines) {
	@fields = split(' ', $x);
	foreach $y (@fields) {
	    @f = split('%', $y);
	    if ($#f == 3) {
		chdir("$f[3]") || last;
		open(DIFF, "-|")
		|| exec $cvs, '-n', '-f', 'diff', '-up',
			    '-r', $f[1], '-r', $f[2], $f[0];
		while (<DIFF>) {
		    chop;

		    # 'cvs diff -u' output  .
		    #
		    #	Index: newlog.pl
		    #	===================================================
		    #	retrieving revision 1.177
		    #	retrieving revision 1.178
		    #	diff -u -r1.177 -r1.178
		    #	...
		    #
		    #  ڵ  line ٷ Ѵ.
		    #
		    #	diff -u -r1.177 -r1.178 newlog.pl
		    #   ...
		    $t = "";
		    if (/^Index: /) {
			s/^Index: //;
			$cur_file = $_;
		    } elsif (/^retrieving / || /^RCS file: / || /^========/) {
			# throw
		    } elsif (/^Binary files /) {
			# VIM syntax color  message
			#   "Binary files .* and .* differ$"
			$t = "Binary files old and new $cur_file differ";
		    } elsif (/^diff /) {
			$t = "$_ $cur_file";
		    } else {
			$t = $_;
		    }
		    if ($t) {
			push(@diff_txt, $t);
		    }
		}
	    }
	}
    }

    # do_status {{{2
    if ($do_status) {
	local(@changed_files);

	@changed_files = ();
	push(@changed_files, &read_logfile("$CHANGED_FILE.$i.$id", ""));
	push(@changed_files, &read_logfile("$ADDED_FILE.$i.$id", ""));
	push(@changed_files, &read_logfile("$REMOVED_FILE.$i.$id", ""));

	if ($debug) {
	    print STDERR "main: pre-sort changed_files = ", join(":", @changed_files), ".\n";
	}
	sort(@changed_files);
	if ($debug) {
	    print STDERR "main: post-sort changed_files = ", join(":", @changed_files), ".\n";
	}

	foreach $dofile (@changed_files) {
	    if ($dofile =~ /\/$/) {
		next;		# ignore the silly "dir" entries
	    }
	    if ($debug) {
		print STDERR "main(): doing status on $dofile\n";
	    }
	    open(STATUS, "-|") || exec 'cvs', '-n', 'status', '-v', $dofile;
	    while (<STATUS>) {
		chop;
		push(@status_txt, $_);
	    }
	}
    }
    # }}}2
}

# Write to the commitlog file {{{1
#
if ($commitlog) {
    &write_commitlog($commitlog, @text);
}

if ($#status_txt >= 0) {
    push(@text, @status_txt);
}
if ($#diff_txt >= 0) {
    push(@text, @diff_txt);
}

# Mailout the notification. {{{1
#
&mail_notification($mailto, @text);

# cleanup {{{1
#
if (! $debug) {
    &cleanup_tmpfiles();
}

# }}}1

exit 0;

# vim:set ts=8 sts=4 sw=4 noet tw=78:
